@@ -290,20 +290,7 @@ deinterleave2_mmx_4 (int16_t *d1, int16_t *d2, int16_t *s_2xn, int n)
 }
 OIL_DEFINE_IMPL_FULL (deinterleave2_mmx_4, deinterleave2_s16, OIL_IMPL_FLAG_MMX);
 
-void
-interleave2_c (int16_t *d_2xn, int16_t *s1, int16_t *s2, int n)
-{
-  int i;
 
-  for(i=0;i<n;i++) {
-    d_2xn[2*i] = s1[i];
-    d_2xn[2*i + 1] = s2[i];
-  }
-}
-OIL_DEFINE_IMPL (interleave2_c, interleave2_s16);
-
-
-//#ifdef ENABLE_BROKEN_IMPLS
 void
 lift_add_mult_shift12_i386_mmx (int16_t *d, int16_t *s1, int16_t *s2,
     int16_t *s3, int16_t *s4, int n)
@@ -349,8 +336,7 @@ lift_add_mult_shift12_i386_mmx (int16_t *d, int16_t *s1, int16_t *s2,
       : "m" (val), "m" (n)
       : "ecx");
 }
-OIL_DEFINE_IMPL (lift_add_mult_shift12_i386_mmx, lift_add_mult_shift12);
-//#endif
+OIL_DEFINE_IMPL_FULL (lift_add_mult_shift12_i386_mmx, lift_add_mult_shift12, OIL_IMPL_FLAG_MMX);
 
 void
 interleave2_mmx (int16_t *d_2xn, int16_t *s1, int16_t *s2, int n)
@@ -691,6 +677,64 @@ mas2_add_s16_mmx (int16_t *d1, int16_t *s1, int16_t *s2, int16_t *s3_2,
 }
 OIL_DEFINE_IMPL_FULL (mas2_add_s16_mmx, mas2_add_s16, OIL_IMPL_FLAG_MMX | OIL_IMPL_FLAG_MMXEXT);
 
+#if 0
+void
+mas2_add_s16_lim_mmx (int16_t *d1, int16_t *s1, int16_t *s2, int16_t *s3_2,
+    int16_t *s4_2, int n)
+{
+  int shift = s4_2[1];
+
+  while (n&3) {
+    int x;
+
+    x = s4_2[0] + s2[0]*s3_2[0] + s2[1]*s3_2[1];
+    x >>= s4_2[1];
+    d1[0] = s1[0] + x;
+
+    d1++;
+    s1++;
+    s2++;
+    n--;
+  }
+  if (n==0) return;
+
+  n>>=2;
+  asm volatile ("\n"
+      "  movzwl 0(%0), %%ecx\n"
+      "  movd %%ecx, %%mm7\n"
+      "  pshufw $0x00, %%mm7, %%mm7\n"
+      "  movzwl 2(%0), %%ecx\n"
+      "  movd %%ecx, %%mm6\n"
+      "  pshufw $0x00, %%mm6, %%mm6\n"
+      "  movzwl 0(%1), %%ecx\n"
+      "  movd %%ecx, %%mm5\n"
+      "  pshufw $0x44, %%mm5, %%mm5\n"
+      :: "r" (s3_2), "r" (s4_2)
+      : "ecx"
+      );
+  asm volatile ("\n"
+      "1:\n"
+      "  movq 0(%2), %%mm0\n"
+      "  paddq 2(%2), %%mm0\n"
+
+      "  movd %4, %%mm4\n"
+      "  psraw %%mm4, %%mm0\n"
+
+      "  paddw 0(%1), %%mm0\n"
+      "  movq %%mm0, 0(%0)\n"
+      "  add $8, %0\n"
+      "  add $8, %1\n"
+      "  add $8, %2\n"
+      "  decl %3\n"
+      "  jnz 1b\n"
+      "  emms\n"
+      : "+r" (d1), "+r" (s1), "+r" (s2), "+r" (n)
+      : "r" (shift)
+      );
+}
+OIL_DEFINE_IMPL_FULL (mas2_add_s16_lim_mmx, mas2_add_s16, OIL_IMPL_FLAG_MMX | OIL_IMPL_FLAG_MMXEXT);
+#endif
+
 void
 mas4_add_s16_mmx (int16_t *d1, int16_t *s1, int16_t *s2, int16_t *s3_4,
     int16_t *s4_2, int n)
@@ -1056,7 +1100,7 @@ mas4_add_s16_pmaddwd (int16_t *d1, int16_t *s1, int16_t *s2, int16_t *s3_2,
       "1:\n"
       "  movq 0(%2), %%mm0\n"
       "  pmaddwd %%mm6, %%mm0\n"
-      "  pshufw $0xee, %%mm0, %%mm1\n"
+      "  pshufw $0xee, %%mm0, %%mm1\n" // 11 10 11 10
       "  paddd %%mm1, %%mm0\n"
       "  paddd %%mm3, %%mm0\n"
       "  psrad %%mm4, %%mm0\n"
@@ -1076,6 +1120,73 @@ mas4_add_s16_pmaddwd (int16_t *d1, int16_t *s1, int16_t *s2, int16_t *s3_2,
 }
 OIL_DEFINE_IMPL_FULL (mas4_add_s16_pmaddwd, mas4_add_s16, OIL_IMPL_FLAG_MMX | OIL_IMPL_FLAG_MMXEXT);
 
+void
+mas4_add_s16_pmaddwd_2 (int16_t *d1, int16_t *s1, int16_t *s2, int16_t *s3_2,
+    int16_t *s4_2, int n)
+{
+  if (n==0) return;
+  asm volatile ("\n"
+      "  movq 0(%0), %%mm6\n"
+      "  movzwl 0(%1), %%ecx\n"
+      "  movd %%ecx, %%mm3\n"
+      "  pshufw $0x44, %%mm3, %%mm3\n" // 01 00 01 00
+      "  movzwl 2(%1), %%ecx\n"
+      "  movd %%ecx, %%mm4\n"
+      :: "r" (s3_2), "r" (s4_2)
+      : "ecx"
+      );
+  if (n&1) {
+    asm volatile ("\n"
+        "  movq 0(%2), %%mm0\n"
+        "  pmaddwd %%mm6, %%mm0\n"
+        "  pshufw $0xee, %%mm0, %%mm1\n" // 11 10 11 10
+        "  paddd %%mm1, %%mm0\n"
+        "  paddd %%mm3, %%mm0\n"
+        "  psrad %%mm4, %%mm0\n"
+        "  movd %%mm0, %%eax\n"
+        "  addw 0(%1), %%ax\n"
+        "  movw %%ax, 0(%0)\n"
+        "  add $2, %0\n"
+        "  add $2, %1\n"
+        "  add $2, %2\n"
+        "  decl %3\n"
+        : "+r" (d1), "+r" (s1), "+r" (s2), "+r" (n)
+        :
+        : "eax"
+        );
+  }
+  n>>=1;
+  asm volatile ("\n"
+      "1:\n"
+      "  movq 0(%2), %%mm0\n"
+      "  pmaddwd %%mm6, %%mm0\n"
+      "  movq 2(%2), %%mm2\n"
+      "  pmaddwd %%mm6, %%mm2\n"
+
+      "  movq %%mm0, %%mm1\n"
+      "  punpckhdq %%mm2, %%mm0\n"
+      "  punpckldq %%mm2, %%mm1\n"
+
+      "  paddd %%mm1, %%mm0\n"
+      "  paddd %%mm3, %%mm0\n"
+      "  psrad %%mm4, %%mm0\n"
+      "  pshufw $0xd8, %%mm0, %%mm0\n" // 11 01 10 00
+
+      "  paddw 0(%1), %%mm0\n"
+      "  movd %%mm0, 0(%0)\n"
+      "  add $4, %0\n"
+      "  add $4, %1\n"
+      "  add $4, %2\n"
+      "  decl %3\n"
+      "  jnz 1b\n"
+      "  emms\n"
+      : "+r" (d1), "+r" (s1), "+r" (s2), "+r" (n)
+      :
+      : "eax"
+      );
+}
+OIL_DEFINE_IMPL_FULL (mas4_add_s16_pmaddwd_2, mas4_add_s16, OIL_IMPL_FLAG_MMX | OIL_IMPL_FLAG_MMXEXT);
+
 void
 mas8_add_s16_pmaddwd (int16_t *d1, int16_t *s1, int16_t *s2, int16_t *s3_2,
     int16_t *s4_2, int n)
@@ -1607,6 +1718,334 @@ mas4_across_add_s16_mmx (int16_t *d, int16_t *s1, int16_t *s2_nx4, int sstr2,
       );
 }
 OIL_DEFINE_IMPL_FULL (mas4_across_add_s16_mmx, mas4_across_add_s16,
-    OIL_IMPL_FLAG_MMX);
+    OIL_IMPL_FLAG_MMX|OIL_IMPL_FLAG_MMXEXT);
 #endif
 
+void
+mas4_across_add_s16_mmx (int16_t *d, int16_t *s1, int16_t *s2_nx4, int sstr2,
+    int16_t *s3_4, int16_t *s4_2, int n)
+{
+  int16_t *s2_nx4_off;
+
+  while (n&3) {
+    int x;
+    int j;
+    x = s4_2[0];
+    for(j=0;j<4;j++){
+      x += OIL_GET(s2_nx4, j*sstr2, int16_t)*s3_4[j];
+    }
+    x >>= s4_2[1];
+    d[0] = s1[0] + x;
+
+    n--;
+    d++;
+    s1++;
+    s2_nx4++;
+  }
+  if (n==0) return;
+
+  s2_nx4_off = OIL_OFFSET(s2_nx4, 3*sstr2);
+
+  n >>= 2;
+  __asm__ __volatile__ ("\n"
+      "  movq 0(%[s3_4]), %%mm0\n"
+      "  pxor %%mm5, %%mm5\n"
+      "  movd 0(%[s4_2]), %%mm5\n"
+      :
+      : [s3_4] "r" (s3_4),
+        [s4_2] "r" (s4_2)
+      );
+
+  __asm__ __volatile__ ("\n"
+      "1:\n"
+      "  pshufw $0x00, %%mm0, %%mm6\n"
+      "  pmullw 0(%[s2_nx4]), %%mm6\n"
+      "  pshufw $0x00, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[s2_nx4]), %%mm3\n"
+      "  movq %%mm6, %%mm7\n"
+      "  punpcklwd %%mm3, %%mm6\n"
+      "  punpckhwd %%mm3, %%mm7\n"
+
+      "  pshufw $0x55, %%mm0, %%mm2\n"
+      "  pmullw 0(%[s2_nx4],%[sstr]), %%mm2\n"
+      "  pshufw $0x55, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[s2_nx4],%[sstr]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  pshufw $0xaa, %%mm0, %%mm2\n"
+      "  pmullw 0(%[s2_nx4],%[sstr],2), %%mm2\n"
+      "  pshufw $0xaa, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[s2_nx4],%[sstr],2), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  pshufw $0xff, %%mm0, %%mm2\n"
+      "  pmullw 0(%[s2_nx4_off]), %%mm2\n"
+      "  pshufw $0xff, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[s2_nx4_off]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  pshufw $0xcc, %%mm5, %%mm1\n"
+      "  paddd %%mm1, %%mm6\n"
+      "  paddd %%mm1, %%mm7\n"
+
+      "  pshufw $0xfd, %%mm5, %%mm1\n"
+      "  psrad %%mm1, %%mm6\n"
+      "  psrad %%mm1, %%mm7\n"
+      "  packssdw %%mm7, %%mm6\n"
+
+      "  paddsw (%[s1]),%%mm6\n"
+      "  movq %%mm6, 0(%[d])\n"
+
+      "  addl $8, %[s2_nx4]\n"
+      "  addl $8, %[s2_nx4_off]\n"
+      "  addl $8, %[s1]\n"
+      "  addl $8, %[d]\n"
+      "  decl %[n]\n"
+      "  jnz 1b\n"
+      "  emms\n"
+      : [s2_nx4] "+r" (s2_nx4),
+        [d] "+r" (d),
+        [s2_nx4_off] "+r" (s2_nx4_off),
+        [n] "+m" (n),
+        [s1] "+r" (s1)
+      : [sstr] "r" (sstr2)
+      );
+}
+OIL_DEFINE_IMPL_FULL (mas4_across_add_s16_mmx, mas4_across_add_s16,
+    OIL_IMPL_FLAG_MMX|OIL_IMPL_FLAG_MMXEXT);
+
+void
+mas8_across_add_s16_mmx (int16_t *d, int16_t *s1, int16_t *s2_nx8, int sstr2,
+    int16_t *s3_8, int16_t *s4_2, int n)
+{
+  int16_t *s2_nx8_off;
+  void *tmp = NULL;
+
+  while (n&3) {
+    int x;
+    int j;
+    x = s4_2[0];
+    for(j=0;j<8;j++){
+      x += OIL_GET(s2_nx8, j*sstr2, int16_t)*s3_8[j];
+    }
+    x >>= s4_2[1];
+    d[0] = s1[0] + x;
+
+    n--;
+    d++;
+    s1++;
+    s2_nx8++;
+  }
+  if (n==0) return;
+
+  s2_nx8_off = OIL_OFFSET(s2_nx8, 7*sstr2);
+
+  n >>= 2;
+  __asm__ __volatile__ ("\n"
+      "  movq 0(%[s3_8]), %%mm0\n"
+      "  pxor %%mm5, %%mm5\n"
+      "  movd 0(%[s4_2]), %%mm5\n"
+      :
+      : [s3_8] "r" (s3_8),
+        [s4_2] "r" (s4_2)
+      );
+
+  __asm__ __volatile__ ("\n"
+      "1:\n"
+      "  movl %[s2_nx8], %[tmp]\n"
+      "  movq 0(%[s3_8]), %%mm0\n"
+
+      "  pshufw $0x00, %%mm0, %%mm6\n"
+      "  pmullw 0(%[tmp]), %%mm6\n"
+      "  pshufw $0x00, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm6, %%mm7\n"
+      "  punpcklwd %%mm3, %%mm6\n"
+      "  punpckhwd %%mm3, %%mm7\n"
+
+      "  addl %[sstr], %[tmp]\n"
+      "  pshufw $0x55, %%mm0, %%mm2\n"
+      "  pmullw 0(%[tmp]), %%mm2\n"
+      "  pshufw $0x55, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  addl %[sstr], %[tmp]\n"
+      "  pshufw $0xaa, %%mm0, %%mm2\n"
+      "  pmullw 0(%[tmp]), %%mm2\n"
+      "  pshufw $0xaa, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  addl %[sstr], %[tmp]\n"
+      "  pshufw $0xff, %%mm0, %%mm2\n"
+      "  pmullw 0(%[tmp]), %%mm2\n"
+      "  pshufw $0xff, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  movq 8(%[s3_8]), %%mm0\n"
+
+      "  addl %[sstr], %[tmp]\n"
+      "  pshufw $0x00, %%mm0, %%mm2\n"
+      "  pmullw 0(%[tmp]), %%mm2\n"
+      "  pshufw $0x00, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  addl %[sstr], %[tmp]\n"
+      "  pshufw $0x55, %%mm0, %%mm2\n"
+      "  pmullw 0(%[tmp]), %%mm2\n"
+      "  pshufw $0x55, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  addl %[sstr], %[tmp]\n"
+      "  pshufw $0xaa, %%mm0, %%mm2\n"
+      "  pmullw 0(%[tmp]), %%mm2\n"
+      "  pshufw $0xaa, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  addl %[sstr], %[tmp]\n"
+      "  pshufw $0xff, %%mm0, %%mm2\n"
+      "  pmullw 0(%[tmp]), %%mm2\n"
+      "  pshufw $0xff, %%mm0, %%mm3\n"
+      "  pmulhw 0(%[tmp]), %%mm3\n"
+      "  movq %%mm2, %%mm4\n"
+      "  punpcklwd %%mm3, %%mm2\n"
+      "  punpckhwd %%mm3, %%mm4\n"
+      "  paddd %%mm2, %%mm6\n"
+      "  paddd %%mm4, %%mm7\n"
+
+      "  pshufw $0xcc, %%mm5, %%mm1\n"
+      "  paddd %%mm1, %%mm6\n"
+      "  paddd %%mm1, %%mm7\n"
+
+      "  pshufw $0xfd, %%mm5, %%mm1\n"
+      "  psrad %%mm1, %%mm6\n"
+      "  psrad %%mm1, %%mm7\n"
+      "  packssdw %%mm7, %%mm6\n"
+
+      "  paddsw (%[s1]),%%mm6\n"
+      "  movq %%mm6, 0(%[d])\n"
+
+      "  addl $8, %[s2_nx8]\n"
+      "  addl $8, %[s1]\n"
+      "  addl $8, %[d]\n"
+      "  decl %[n]\n"
+      "  jnz 1b\n"
+      "  emms\n"
+      : [s2_nx8] "+r" (s2_nx8),
+        [tmp] "+r" (tmp),
+        [s3_8] "+r" (s3_8),
+        [d] "+r" (d),
+        [n] "+m" (n),
+        [s1] "+r" (s1)
+      : [sstr] "m" (sstr2)
+      );
+}
+OIL_DEFINE_IMPL_FULL (mas8_across_add_s16_mmx, mas8_across_add_s16,
+    OIL_IMPL_FLAG_MMX|OIL_IMPL_FLAG_MMXEXT);
+
+void
+lshift_s16_mmx(int16_t *d1, int16_t *s1, int16_t *s3_1, int n)
+{
+  while (n&3) {
+    d1[0] = s1[0]<<s3_1[0];
+    d1++;
+    s1++;
+    n--;
+  }
+  n >>= 2;
+  __asm__ __volatile__ ("\n"
+      "  movzwl 0(%[s3_1]), %%ecx\n"
+      "  movd %%ecx, %%mm1\n"
+      "1:\n"
+      "  movq 0(%[s1]), %%mm0\n"
+      "  psllw %%mm1, %%mm0\n"
+      "  movq %%mm0, 0(%[d1])\n"
+      "  add $8, %[d1]\n"
+      "  add $8, %[s1]\n"
+      "  decl %[n]\n"
+      "  jnz 1b\n"
+      "  emms"
+      : [d1] "+r" (d1),
+        [s1] "+r" (s1),
+        [n] "+r" (n)
+      : [s3_1] "r" (s3_1)
+      : "ecx");
+}
+OIL_DEFINE_IMPL_FULL (lshift_s16_mmx, lshift_s16, OIL_IMPL_FLAG_MMX);
+
+void
+lshift_s16_mmx_2(int16_t *d1, int16_t *s1, int16_t *s3_1, int n)
+{
+  while (n&7) {
+    d1[0] = s1[0]<<s3_1[0];
+    d1++;
+    s1++;
+    n--;
+  }
+  n >>= 3;
+  if (n == 0) return;
+  __asm__ __volatile__ ("\n"
+      "  movzwl 0(%[s3_1]), %%ecx\n"
+      "  movd %%ecx, %%mm1\n"
+      "1:\n"
+      "  movq 0(%[s1]), %%mm0\n"
+      "  psllw %%mm1, %%mm0\n"
+      "  movq %%mm0, 0(%[d1])\n"
+      "  movq 8(%[s1]), %%mm0\n"
+      "  psllw %%mm1, %%mm0\n"
+      "  movq %%mm0, 8(%[d1])\n"
+      "  add $16, %[d1]\n"
+      "  add $16, %[s1]\n"
+      "  decl %[n]\n"
+      "  jnz 1b\n"
+      "  emms"
+      : [d1] "+r" (d1),
+        [s1] "+r" (s1),
+        [n] "+r" (n)
+      : [s3_1] "r" (s3_1)
+      : "ecx");
+}
+OIL_DEFINE_IMPL_FULL (lshift_s16_mmx_2, lshift_s16, OIL_IMPL_FLAG_MMX);
+
+
